#include "wire.h"
#include "swd.h"

//addressed for APnDP = 0
#define ADDR_IDCODE_R	0
#define ADDR_ABORT_W	0
#define ADDR_STAT_R	1
#define ADDR_CTRL_W	1
#define ADDR_TARGETID	1
#define ADDR_SELECT_W	2
#define ADDR_RDBUFF_R	3


#define AP_CLASS_MEMORY		0x08
#define AP_CONN_TYPE_AHB	0x01
#define AP_CONN_TYPE_APB	0x02
#define AP_CONN_TYPE_AXI	0x04


#ifdef AVR
#include <avr/io.h>
#define mNumAps	OCR0A
#else
static uint8_t mNumAps;
#endif

static inline bool swdWireBusReadInt(uint1_t ap, uint8_t a23, unaligned_uint32_t *valP)
{
	uint8_t ret;
	uint8_t tryNo = 0;
	
	do {
		ret = llWireBusRead(ap, a23, valP);
	} while ((ret &~ (uint8_t)BUS_DATA_PAR_ERR) == BUS_SWD_WAIT && ++tryNo);
	
	return BUS_SWD_ACK == ret;
}

static inline bool swdWireBusWriteInt(uint1_t ap, uint8_t a23, uint32_t val)
{
	uint8_t ret;
	uint8_t tryNo = 0;
	
	do {
		ret = llWireBusWrite(ap, a23, val);
	} while ((ret &~ (uint8_t)BUS_DATA_PAR_ERR) == BUS_SWD_WAIT && ++tryNo);
	
	return BUS_SWD_ACK == ret;
}

bool swdWireBusRead(uint1_t ap, uint8_t a23, unaligned_uint32_t *valP)
{
	return swdWireBusReadInt(ap, a23, valP);
}

bool swdWireBusWrite(uint1_t ap, uint8_t a23, uint32_t val)
{
	return swdWireBusWriteInt(ap, a23, val);
}

bool swdReadIdcode(unaligned_uint32_t *valP)
{
	return swdWireBusReadInt(0, ADDR_IDCODE_R, valP);
}

bool swdReadStat(unaligned_uint32_t *valP)
{
	return swdWireBusReadInt(0, ADDR_STAT_R, valP);
}

bool swdReadRdbuf(unaligned_uint32_t *valP)
{
	return swdWireBusReadInt(0, ADDR_RDBUFF_R, valP);
}

bool swdWriteAbort(uint32_t val)
{
	return swdWireBusWriteInt(0, ADDR_ABORT_W, val);
}

bool swdWriteSelect(uint32_t val)
{
	return swdWireBusWriteInt(0, ADDR_SELECT_W, val);
}

bool swdWriteCtrl(uint32_t val)
{
	return swdWireBusWriteInt(0, ADDR_CTRL_W, val);
}

uint8_t swdAttach(uint8_t *mIdxAhb, unaligned_uint32_t *targetId /* or 0 if reg not existent */)
{
	uint32_t val;
	bool ahbFound = false;
	uint8_t ahbIdx, dapVer, i;

	mNumAps = 0;

	if (!swdReadIdcode(WRAP_UNALIGNED_POINTER_32(&val)))
		return SWD_ATTACH_IDCODE_READ_FAIL;

	dapVer = ((uint8_t)(((uint16_t)val) >> 12)) & 0x0F;
	if (!dapVer) {
//		pr(PSTR("DAPv0 not supported\n"));
		return SWD_ATTACH_DAP_V0;
	}

	if (!swdWriteAbort(0x1F)) {
//		pr(PSTR("aborts clear failed\n"));
		return SWD_ATTACH_REG_WRITE_FAIL;
	}
	
	if (!swdWriteSelect(0)) {
//		pr(PSTR("select failed\n"));
		return SWD_ATTACH_REG_WRITE_FAIL;
	}
	
	//power up cpu and debug infrastructure
	if (!swdWriteCtrl(0x50000000UL)) {
//		pr(PSTR("power up failed\n"));
		return SWD_ATTACH_REG_WRITE_FAIL;
	}
	
	//wait for powerup
	for (i = 0; i < 16; i++) {
	
		if (!swdReadStat(WRAP_UNALIGNED_POINTER_32(&val)))
			return SWD_ATTACH_REG_READ_FAIL;
	
		if ((uint8_t)(val >> 28) == 0xF)
			goto powerup_ok;
	}
	
//	pr(PSTR("CTRL/STAT = 0x%08lX\n"), (unsigned long)val);
	return SWD_ATTACH_CTRL_STAT_VAL_ERR;
	
powerup_ok:

	if (dapVer < 2) {
		UNALIGNED(targetId) = 0;
	}
	else {
		if (!swdWriteSelect(0x02)) {
//			pr(PSTR("Failed to select DPBANK 2\n"));
			return SWD_ATTACH_REG_WRITE_FAIL;
		}
		if (!swdWireBusReadInt(0, ADDR_TARGETID, targetId)) {
//			pr(PSTR("Failed to select TARGETID 2\n"));
			return SWD_ATTACH_REG_WRITE_FAIL;
		}
	}
	
	mNumAps = 0;
	do {
		uint8_t cls, var, typ;
		
		if (!swdWriteSelect((((uint32_t)mNumAps) << 24) | 0xF0)) {
//			pr(PSTR("Failed to select [%3u]\n"), mNumAps);
			return SWD_ATTACH_REG_WRITE_FAIL;
		}
		if (!swdWireBusReadInt(1, 3, WRAP_UNALIGNED_POINTER_32(&val))) {
//			pr(PSTR("Failed to enqueue [%3u].IDR read\n"), mNumAps);
			return SWD_ATTACH_REG_READ_FAIL;
		}
		if (!swdReadRdbuf(WRAP_UNALIGNED_POINTER_32(&val))){
//			pr(PSTR("Failed to capture [%3u].IDR read\n"), mNumAps);
			return SWD_ATTACH_RDBUF_READ_FAIL;
		}
		if (!val)
			break;
		
		//skip all APs that are not by "ARM"
		if ((val & 0x0FFEFF00) == 0x04760000) {
			cls = ((uint8_t)(val >> 13)) & 0x0f;
			var = ((uint8_t)val) >> 4;
			typ = ((uint8_t)val) & 0x0F;
	//		pr(PSTR("[%3u].IDR = 0x%08lX (v %d, var %2d typ %2d)\n"), mNumAps, val, cls, var, typ);
			
			if (cls == AP_CLASS_MEMORY && typ == AP_CONN_TYPE_AHB) {
				ahbIdx = mNumAps;
				ahbFound = true;
			}
		}
		else {
		//		pr(PSTR("[%3u].IDR not understandable 0x%08lX\n"), mNumAps, val);
		}
		
		mNumAps++;
	} while(1);
	
	if (!ahbFound) {
//		pr(PSTR("AHB not found\n"));
		return SWD_ATTACH_AHB_NOT_FOUND;
	}

//	pr(PSTR("Found %u AP(s)\n"), mNumAps);

	if (mIdxAhb)
		*mIdxAhb = ahbIdx;

	return SWD_ATTACH_RET_OK;
}